#ifdef GRAPHICS_H
#error Double include.
#endif
#define GRAPHICS_H

enum {
    MAX_JBUF      = 16,
    MAX_ANITYPE   = 255,
    #ifdef NO_ANIM
        MAX_ANIM  = 1,
    #else
        MAX_ANIM  = 2500,
    #endif
};

enum TextFlags {
    Txt_Left       = 0x00,
    Txt_Center     = 0x01,
    Txt_Right      = 0x02,
    Txt_Underline  = 0x04,
    Txt_Justify    = 0x08,
    Txt_AltFont    = 0x10,
    Txt_Hashed     = 0x20,
    Txt_WordBreak  = 0x40,
    Txt_Border     = 0x80
};

enum AnimFlags {
    Ani_Atk_Rewind = 0x01,
    Ani_Center     = 0x10,
    Ani_3_4ths     = 0x20,
};

typedef void SFUNC (class CSprite*);

#ifdef _DEBUG
    extern void error_hook (void);
#else
    #define error_hook()
#endif

struct my_error_mgr {
    jpeg_error_mgr pub;
    jmp_buf        setjmp_buffer;
};

struct CAniType {
    unsigned number,
             scheme,
             flags;
};

class CSprite {
    public:                CSprite         (void);
                           ~CSprite        (void);
    CSprite*               AddToList       (IDirectDrawSurface*);
    bool                   Load            (const char*);
    bool                   Create          (unsigned, unsigned, char*, bool);
    void inline            Blit            (int, int, int, int, int, int);
    void inline            BlitTr          (int, int, int, int, int, int);
    void inline            BlitTo          (int, int, int, int, int, int, CSprite *);
    void                   BlitTiled       (int, int, int, int, int, int, int, int);
    void                   Destroy         (void);

    CSprite                *next;
    bool                   setcolorkey;
    char                   *bmpbuf;
    IDirectDrawSurface     *surface;
    SFUNC                  *sprfunc,
                           *postfunc;
    unsigned               image,
                           xsize,
                           ysize;
};

class CAnimation {
    public:                CAnimation      (void);
    unsigned               GetFrame        (unsigned, unsigned);
    void                   Display         (int, int, unsigned, unsigned);
    void                   DisplayThumb    (int, int);
    void                   Load            (unsigned, bool);
    unsigned               startframe      [6 * 8],
                           numframes       [6 * 8],
                           loaded,
                           flags,
                           refcount;
};

class CAnimList { public:
    CAnimation             *anim;
    unsigned               x,
                           y,
                           pose,
                           frame;
    CAnimList              *next;
};

extern CAnimation Anims    [];
extern CAniType   anitypes [];

class CScreen: public CDialog {
    public:                CScreen         (void);
    virtual                ~CScreen        (void);
    bool                   Init            (unsigned, unsigned, unsigned, unsigned);
    bool                   Restore         (void);
    void                   Destroy         (void);
    bool                   CreateSurface   (unsigned, unsigned, IDirectDrawSurface**, bool);
    void                   Flip            (void);
    void                   FlipCopy        (void);
    CSprite*               LoadSprite      (unsigned, SFUNC* func = NULL, SFUNC* pfunc = NULL);
    bool                   LoadJpeg        (unsigned, bool vidmem = true, unsigned xsz = 0, unsigned ysz = 0);
    unsigned               MergeJpeg       (unsigned, unsigned, unsigned, unsigned, unsigned, unsigned, unsigned);
    unsigned               TextWidth       (const char*, bool);
    void                   PutTextBox      (int, int, int, unsigned, const char*);
    void                   PutTextBox      (int, int, int, unsigned, unsigned);
    // the below doesn't function in windowed mode:
    void                   PutRect         (unsigned, unsigned, unsigned, unsigned, unsigned);
    void                   PutGray         (int, int, int, int, unsigned);
    void                   PutEdge         (unsigned, unsigned, unsigned, unsigned, unsigned);
    void                   PutLine         (unsigned, unsigned, unsigned, unsigned, unsigned);
    void                   ClipLine        (int, int, int, int, unsigned);
    void                   PutCircle       (unsigned, unsigned, unsigned, unsigned);
    void inline            PsetXfer2_16    (unsigned, unsigned);
    void inline            PsetXfer2_32    (unsigned, unsigned);
    void inline            PsetClip_16     (unsigned, unsigned, short);
    void inline            PsetClip_32     (unsigned, unsigned, unsigned);
    void inline            PMemset_16      (unsigned, short);
    void inline            PMemset2_16     (unsigned, unsigned);
    void inline            PMemset_32      (unsigned, unsigned);
    void inline            PsetXfer_32     (unsigned, unsigned);
    /*
    void inline            Pset2_16        (unsigned, unsigned, unsigned);
    void inline            PsetXfer_16     (unsigned, unsigned);
    void inline            PsetXferFm_16   (unsigned, unsigned, unsigned, unsigned);
    void inline            PsetXferFm_32   (unsigned, unsigned, unsigned, unsigned);
    */
    unsigned inline        Pget_16         (unsigned, unsigned);
    void inline            Pset_16         (unsigned, unsigned, short);
    unsigned inline        Pget_32         (unsigned, unsigned);
    void inline            Pset_32         (unsigned, unsigned, unsigned);
    unsigned inline        StepX_16        (void);
    unsigned inline        StepX_32        (void);
    unsigned inline        StepY           (void);
    unsigned inline        GetColor        (unsigned);
    void                   Lock            (CSprite*);
    void                   Unlock          (void);
    // the below doesn't work in windowed mode
    void                   LockDual        (CSprite*, CSprite*);
    /*
    void                   SaveBitmap      (CSprite*, const char*);
    */
    void                   WaitVSync       (void);
    void                   SaveScreenshot  (void);
    void                   ConvertSprites  (unsigned);
    void                   MakeSprFile     (const char*);
    void                   AnimStart       (void);
    void                   AnimAdd         (CAnimation*, unsigned, unsigned, unsigned, unsigned);
    void                   AnimShow        (void);

    int                    hires,
                           fontdata        [256];
    unsigned               winxsize,
                           winysize,
                           mousex,
                           mousey,
                           mousecursorx,
                           mousecursory,
                           mousecursorsz,
                           animstate,
                           dispstate,
                           fade,
                           offset_bmp,
                           offset_spr,
                           offset_jpeg,
                           last_jpeg,
                           nextsprite,
                           videomem,
                           winbpp;
    bool                   fullscreen,
                           usewindow,
                           doflip;
    IDirectDrawSurface     *lpDDSPrimary,
                           *buffer;
    SDL_Surface            *screen;
    CSprite                sprlist,
                           animsprites     [MAX_ANIM],
                           *font,
                           *jpeg,
                           *back,
                           *front,
                           *fadebuf;

    private:
    bool                   BaseInit        (unsigned, unsigned, unsigned, unsigned);
    void                   ProgBar         (bool);
    const char*            DXError         (HRESULT);

    IDirectDrawSurface     *locksurf,
                           *locksurf2;
    IDirectDrawGammaControl*gamma;
    DDGAMMARAMP            gammaramp;
    CAnimList              *animlist,
                           *junkfirst,
                           *junklast;
    CSprite                *jpegbuf        [MAX_JBUF];
    unsigned               jpeg_cache      [MAX_JBUF],
                           jpeg_cycle,
                           jpeg_merge1,
                           jpeg_merge2;
    HICON                  m_hIcon;
    LPDIRECTDRAW           lpDD;
    DDBLTFX                bltfx;
    DDSURFACEDESC          ddsd,
                           ddsd2;
    jpeg_decompress_struct cinfo;
    my_error_mgr           jerr;

    // Dialog Data
    //{{AFX_DATA(CScreen)
    //}}AFX_DATA
    // Overrides
    //{{AFX_VIRTUAL(CScreen)
    //}}AFX_VIRTUAL
    // Implementation

    //{{AFX_MSG(CScreen)
    virtual BOOL    OnInitDialog    (void);
    afx_msg void    OnPaint         (void);
    afx_msg HCURSOR OnQueryDragIcon (void);
    afx_msg void    OnActivate      (UINT, CWnd*, BOOL);
    //}}AFX_MSG
    DECLARE_MESSAGE_MAP()
};

extern CScreen Screen;

#ifndef USE_ALLEGRO

void inline CScreen::PsetClip_16 (unsigned x, unsigned y, short color) {
    if (x >= 0 && x < winxsize
    &&  y >= 0 && y < winysize) Pset_16 (x, y, color);
}

void inline CScreen::PsetClip_32 (unsigned x, unsigned y, unsigned color) {
    if (x >= 0 && x < winxsize
    &&  y >= 0 && y < winysize) Pset_32 (x, y, color);
}

#ifndef LINUX

unsigned inline CScreen::GetColor (unsigned n) {
    return ((((n & 0xf80000) >> 8) | ((n & 0x00fc00) >> 5) | ((n & 0x0000f8) >> 3)));
}

void inline CScreen::Pset_16 (unsigned x, unsigned y, short color) {
    *((short*) ddsd.lpSurface + x + (ddsd.lPitch * y)) = color;
}

/* only used in SaveBitmap and pixelizer */
    unsigned inline CScreen::Pget_16 (unsigned x, unsigned y) {
        return *((short*) ddsd.lpSurface + x + (ddsd.lPitch * y));
    }


/* unused *
    void inline CScreen::Pset2_16 (unsigned x, unsigned y, unsigned color) {
        *(int*)((short*) ddsd.lpSurface + x + (ddsd.lPitch * y)) = color;
    }

    void inline CScreen::PsetXfer_16 (unsigned x, unsigned y) {
        *((short*) ddsd2.lpSurface + x + (ddsd2.lPitch * y)) =
        *((short*) ddsd. lpSurface + x + (ddsd. lPitch * y));
    }

    void inline CScreen::PsetXferFm_16 (unsigned x, unsigned y, unsigned x2, unsigned y2) {
        *((short*) ddsd. lpSurface + x  + (ddsd. lPitch * y)) =
        *((short*) ddsd2.lpSurface + x2 + (ddsd2.lPitch * y2));
    }
*/

void inline CScreen::PsetXfer2_16 (unsigned x, unsigned y) {
    *(int*)((short*) ddsd2.lpSurface + x + (ddsd2.lPitch * y)) =
    *(int*)((short*) ddsd. lpSurface + x + (ddsd. lPitch * y));
}

void inline CScreen::PMemset_16 (unsigned pos, short color) {
    *((short*) ddsd.lpSurface + pos) = color;
}

void inline CScreen::PMemset2_16 (unsigned pos, unsigned color) {
    *(int*)((short*) ddsd.lpSurface + pos) = color;
}

unsigned inline CScreen::StepX_16 (void) { return 1;           }
unsigned inline CScreen::StepX_32 (void) { return 1;           }
unsigned inline CScreen::StepY    (void) { return ddsd.lPitch; }

void     inline CScreen::Pset_32       (unsigned, unsigned, unsigned)           { error_hook (); }
unsigned inline CScreen::Pget_32       (unsigned, unsigned)                     { error_hook (); return 0; }
void     inline CScreen::PMemset_32    (unsigned, unsigned)                     { error_hook (); }
void     inline CScreen::PsetXfer_32   (unsigned, unsigned)                     { error_hook (); }
void     inline CScreen::PsetXfer2_32  (unsigned, unsigned)                     { error_hook (); }
/*
void     inline CScreen::PsetXferFm_32 (unsigned, unsigned, unsigned, unsigned) { error_hook (); }
*/

void inline CSprite::Blit (int tx, int ty, int sx, int sy, int ax, int ay) {
    if (ax < 0 || ay < 0) error_hook ();
    if (this == NULL) error_hook ();
    Screen.buffer->BltFast (tx, ty, surface, CRect (sx, sy, sx + ax, sy + ay), DDBLTFAST_NOCOLORKEY  | DDBLTFAST_WAIT);
}

void inline CSprite::BlitTo (int tx, int ty, int sx, int sy, int ax, int ay, CSprite *dest) {
    if (ax < 0 || ay < 0) error_hook ();
    if (this == NULL || dest == NULL) error_hook ();
    dest->surface->BltFast (tx, ty, surface, CRect (sx, sy, sx + ax, sy + ay), DDBLTFAST_NOCOLORKEY  | DDBLTFAST_WAIT);
}

void inline CSprite::BlitTr (int tx, int ty, int sx, int sy, int ax, int ay) {
    if (ax < 0 || ay < 0) error_hook ();
    if (this == NULL) error_hook ();
    Screen.buffer->BltFast (tx, ty, surface, CRect (sx, sy, sx + ax, sy + ay), DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT);
}

#else

unsigned inline CScreen::GetColor (unsigned color) {
    return SDL_MapRGB (screen->format, (color & 0xff0000) >> 16,
                                       (color & 0x00ff00) >> 8,
                                        color & 0x0000ff);
}

void inline CScreen::Pset_16 (unsigned x, unsigned y, short color) {
    *(Uint16*)((Uint8*) locksurf->pixels + (y * locksurf->pitch) + 2 * x) = (Uint16) color;
}

void inline CScreen::Pset2_16 (unsigned x, unsigned y, unsigned color) {
    *(Uint32*)((Uint8*) locksurf->pixels + (y * locksurf->pitch) + 2 * x) = color;
}

unsigned inline CScreen::Pget_16 (unsigned x, unsigned y) {
    return *(Uint16*)((Uint8*) locksurf->pixels + (y * locksurf->pitch) + 2 * x);
}

void inline CScreen::PsetXfer_16 (unsigned x, unsigned y) {
    *(Uint16*)((Uint8*) locksurf2->pixels + (y * locksurf2->pitch) + 2 * x) =
    *(Uint16*)((Uint8*) locksurf ->pixels + (y * locksurf ->pitch) + 2 * x);
}

void inline CScreen::PsetXferFm_16 (unsigned x, unsigned y, unsigned x2, unsigned y2) {
    *(Uint16*)((Uint8*) locksurf ->pixels + (y  * locksurf ->pitch) + 2 * x) =
    *(Uint16*)((Uint8*) locksurf2->pixels + (y2 * locksurf2->pitch) + 2 * x2);
}

void inline CScreen::PsetXfer2_16 (unsigned x, unsigned y) {
    *(Uint32*)(((Uint8*)(locksurf2->pixels)) + (y * locksurf2->pitch) + 2 * x) =
    *(Uint32*)(((Uint8*)(locksurf ->pixels)) + (y * locksurf ->pitch) + 2 * x);
}

void inline CScreen::Pset_32 (unsigned x, unsigned y, unsigned color) {
    *(Uint32*)((Uint8*) locksurf->pixels + (y * locksurf->pitch) + 4 * x) = color;
}

unsigned inline CScreen::Pget_32 (unsigned x, unsigned y) {
    return *(Uint32*)((Uint8*) locksurf->pixels + (y * locksurf->pitch) + 4 * x);
}

void inline CScreen::PsetXfer_32 (unsigned x, unsigned y) {
    *(Uint32*)((Uint8*) locksurf2->pixels + (y * locksurf2->pitch) + 4 * x) =
    *(Uint32*)((Uint8*) locksurf ->pixels + (y * locksurf ->pitch) + 4 * x);
}

void inline CScreen::PsetXferFm_32 (unsigned x, unsigned y, unsigned x2, unsigned y2) {
    *(Uint32*)((Uint8*) locksurf ->pixels + (y  * locksurf ->pitch) + 4 * x) =
    *(Uint32*)((Uint8*) locksurf2->pixels + (y2 * locksurf2->pitch) + 4 * x2);
}

void inline CScreen::PsetXfer2_32 (unsigned x, unsigned y) {
    *(Uint32*)(((Uint8*)(locksurf2->pixels)) + (y * locksurf2->pitch) + 4 * x) =
    *(Uint32*)(((Uint8*)(locksurf ->pixels)) + (y * locksurf ->pitch) + 4 * x);

    *(Uint32*)(((Uint8*)(locksurf2->pixels)) + (y * locksurf2->pitch) + 4 * x + 4) =
    *(Uint32*)(((Uint8*)(locksurf ->pixels)) + (y * locksurf ->pitch) + 4 * x + 4);
}

void inline CScreen::PMemset_16 (unsigned pos, short color) {
    *(Uint16*)((Uint8*) locksurf->pixels + pos) = (Uint16) color;
}

void inline CScreen::PMemset2_16 (unsigned pos, unsigned color) {
    *(Uint32*)((Uint8*) locksurf->pixels + pos) = color;
}

void inline CScreen::PMemset_32 (unsigned pos, unsigned color) {
    *(Uint32*)((Uint8*) locksurf->pixels + pos) = color;
}

unsigned inline CScreen::StepX_16 (void) { return 2;               }
unsigned inline CScreen::StepX_32 (void) { return 4;               }
unsigned inline CScreen::StepY    (void) { return locksurf->pitch; }

void inline CSprite::Blit (int tx, int ty, int sx, int sy, int ax, int ay) {
    if (setcolorkey) {
        SDL_SetColorKey (surface, 0, 0);
        setcolorkey = false;
    }
    SDL_Rect srcrect = { sx, sy, ax, ay };
    SDL_Rect dstrect = { tx, ty,  0,  0 };
    SDL_BlitSurface (surface, &srcrect, Screen.screen, &dstrect);
}

void inline CSprite::BlitTo (int tx, int ty, int sx, int sy, int ax, int ay, CSprite *dest) {
    if (setcolorkey) {
        SDL_SetColorKey (surface, 0, 0);
        setcolorkey = false;
    }
    SDL_Rect srcrect = { sx, sy, ax, ay };
    SDL_Rect dstrect = { tx, ty,  0,  0 };
    SDL_BlitSurface (surface, &srcrect, dest->surface, &dstrect);
}

void inline CSprite::BlitTr (int tx, int ty, int sx, int sy, int ax, int ay) {
    if (setcolorkey == false) {
        SDL_SetColorKey (surface, SDL_SRCCOLORKEY, 0);
        setcolorkey = true;
    }
    SDL_Rect srcrect = { sx, sy, ax, ay };
    SDL_Rect dstrect = { tx, ty,  0,  0 };
    SDL_BlitSurface (surface, &srcrect, Screen.screen, &dstrect);
}

#endif
#else

unsigned inline CScreen::GetColor (unsigned n) {
    return makecol16 ((n & 0xff0000) >> 16, (n & 0xff00) >> 8, (n & 0xff));
}

unsigned inline CScreen::Pget_16 (unsigned x, unsigned y) {
    return getpixel (locksurf, x, y);
}

unsigned inline CScreen::Pget_32 (unsigned x, unsigned y) {
    return getpixel (locksurf, x, y);
}

void inline CSprite::Blit (int tx, int ty, int sx, int sy, int ax, int ay) {
    if (ax < 0 || ay < 0) error_hook ();
    if (this == NULL) error_hook ();
    rectfill (Screen.buffer, tx, ty, tx+ax-1, ty+ay-1, 0);
    masked_blit (surface, Screen.buffer, sx, sy, tx, ty, ax, ay);
}

void inline CSprite::BlitTo (int tx, int ty, int sx, int sy, int ax, int ay, CSprite *dest) {
    if (ax < 0 || ay < 0) error_hook ();
    if (this == NULL || dest == NULL) error_hook ();
    blit (surface, dest->surface, sx, sy, tx, ty, ax, ay);
}

void inline CSprite::BlitTr (int tx, int ty, int sx, int sy, int ax, int ay) {
    if (ax < 0 || ay < 0) error_hook ();
    if (this == NULL) error_hook ();
    masked_blit (surface, Screen.buffer, sx, sy, tx, ty, ax, ay);
}

void inline CScreen::PutLine (unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned color) {
    line (buffer, x1, y1, x2, y2, color);
}

void inline CScreen::ClipLine (int x1, int y1, int x2, int y2, unsigned color) {
    line (buffer, x1, y1, x2, y2, color);
}

void inline CScreen::PutRect (unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned color) {
    rectfill (buffer, x1, y1, x2, y2, color);
}

void inline CScreen::PutEdge (unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned color) {
    rect (buffer, x1, y1, x2, y2, color);
}

void inline CScreen::PutCircle (unsigned cx, unsigned cy, unsigned cr, unsigned color) {
    circle (buffer, cx, cy, cr, color);
}

#endif

//{{AFX_INSERT_LOCATION}}
