
bool ReadScript (struct CParser *this, unsigned offset, unsigned scrnum) {
    FILE     *file;
    unsigned i, j;
    if (script) free (script);
    i = tracelevel;
    j = number;
    memset (this, 0, sizeof (struct CParser));
    tracelevel = i;
    number     = j;
    for (i=Str_W_Last + 1; i<Str_SysLast; i++) if (StrTable[i]) {
        AddVar (this, i, P_NewFunc);
        ((struct CVar*) varptr)->arraysize = i;
        ((struct CVar*) varptr)->value     = StrTable[i][0] - '0';
    }
    
    /* read the file */
    file = fopen ("script.txt", "rb");
    if (file == NULL) return false;
    fseek (file, 0, SEEK_END);
    scrlen = ftell (file);
    rewind (file);
    script = malloc (scrlen);
    if (script == NULL) {
        printf ("Memory error\n");
        return false;
    }
    fread (script, scrlen, 1, file); 
    if (ferror (file)) return false;
    fclose (file);
    
    /* Convert to lowercase */
    for (i=0, j=0; i<scrlen; i++) {
        if (j == 0) {
            if (script[i] >= 'A' && script[i] <= 'Z') script[i] += 'a' - 'A';
            if (script[i] == '"') j = 1;
        } else {
            if (script[i] == '"' || script[i] < ' ') j = 0;
        }
    }
    MakeByteCode (this);
    
    /* Initial pass over the script to weed out any
     * lexical errors. This catches most mistakes, but
     * not things like overflow or eternal recursion.
     */
    while (1) {
        i = Process (this);
        if (i == Str_S_None) {
            if (mainpos) {
                skip   = 0;
                end    = 0;
                pos    = mainpos;
                action = Str_W_Main;
                PushScope (this, P_Command);
                break;
            }
            i = Str_S_NoMain;
        }
        if (i) {
            PrintError (this, i);
            return false;
        }
    }
    return true;
}


/* Print error message and line number (approx) */
void PrintError (struct CParser *this, unsigned errnum) {
    char     b1 [1024];
    unsigned i, j=1, k=0;
    if (script == NULL) b1[1] = 0;
    else for (i=0; script[i]; i++) {
        if (script[i] == 0x0a) {
            j++;
            k = 0;
            if (j > curline) break;
        } else if (script[i] >= ' ') {
            b1[k] = script[i];
            if (k < 1023) k++;
            b1[k] = 0;
        }
    }
    printf ("%s%3.3d: %s\n%s", StrTable[Str_ErrParse], j, b1, StrTable[errnum]);
}

/* Add a new variable of a given identifier tref;
 * note that functions are a type of variable.
 */
void AddVar (struct CParser *this, unsigned tref, unsigned type) {
    struct CVar *temp,
                *t2;
    if (tref == 0) return;
    temp = malloc (sizeof (struct CVar));
    if (temp == NULL) {
        printf ("memory error\n");
        exit (1);
    }
    memset (temp, 0, sizeof (struct CVar));
    temp->tokenref = tref;
    varptr   = &temp->value;
    switch (type) {
        case P_NewFunc:  temp->scope   = pos;
                         temp->is_func = true;
                         break;
        case P_NewVar:   temp->scope   = scopelevel - 1;
                         temp->is_func = false;
                         break;
    }
    
    /* Insert into the variable list, BEFORE any other
     * variables of the same name (scope overloading)
     */
    if (varlist == NULL) {
        varlist = temp;
    } else {
        for (t2 = varlist; t2 != NULL; t2 = t2->next) {
            if (tref <= t2->tokenref) {
                temp->next = t2;
                if (t2 == varlist) {
                    varlist = temp;
                } else {
                    for (t2 = varlist; t2->next != temp->next; t2 = t2->next);
                    t2->next = temp;
                }
                break;
            }
            if (t2->next == NULL) {
                t2->next   = temp;
                temp->next = NULL;
                break;
            }
        }
    }
    Trace (this, 2, StrTable[Str_S_AddVar], tref, temp);
}

/* Remove all variables in a given scope. Used when the scope
 * ends. This is useful for having multiple variables with the same
 * name in different scopes, or in recursive functions */
void KillVars (struct CParser *this) {
    struct CVar *temp,
                *prev;
    if (varlist) {
        while (varlist != NULL && varlist->scope > scopelevel && varlist->is_func == false) {
            temp = varlist;
            varlist = varlist->next;
            Trace (this, 2, StrTable[Str_S_KillVar], temp->tokenref, temp);
            if (temp->arraydata) free (temp->arraydata);
            free (temp);
        }
    }
    if (varlist) {
        prev = varlist;
        temp = prev->next;
        while (temp != NULL) {
            if (temp->scope > scopelevel && temp->is_func == false) {
                prev->next = temp->next;
                Trace (this, 2, StrTable[Str_S_KillVar], temp->tokenref, temp);
                if (temp->arraydata) free (temp->arraydata);
                free (temp);
                temp = prev->next;
            } else {
                prev = temp;
                temp = temp->next;
            }
        }
    }
}

void PopScope (struct CParser *this) {
    unsigned i, j;
    if (scopelevel == 0) {
        /* This shouldn't ever occur - pop upon an empty stack */
        Trace (this, 0, "Logical error???\n", 0, " ");
        exit (1);
    } else {
        scopelevel--;
        if (skip != Skip_Break) skip = scopestack[scopelevel].skip;
        retpos = scopestack[scopelevel].retpos;
        if (scopestack[scopelevel].action >= Str_W_Assign && scopestack[scopelevel].action <= Str_W_Dec) {
            if (scopestack[scopelevel].varptr == NULL) {
		        Trace (this, 0, "Logical error???\n", 0, " ");
				exit (1);
            }
            /* Variable assignments: = += ++ *= etc */
            switch (scopestack[scopelevel].action) {
                case Str_W_Assign:                                                                           break;
                case Str_W_AsgAdd:  value = *scopestack[scopelevel].varptr +  value; break;
                case Str_W_AsgSub:  value = *scopestack[scopelevel].varptr -  value; break;
                case Str_W_AsgMult: value = *scopestack[scopelevel].varptr *  value; break;
                case Str_W_AsgAnd:  value = *scopestack[scopelevel].varptr &  value; break;
                case Str_W_AsgOr:   value = *scopestack[scopelevel].varptr |  value; break;
                case Str_W_AsgXor:  value = *scopestack[scopelevel].varptr ^  value; break;
                case Str_W_AsgLSh:  value = *scopestack[scopelevel].varptr << value; break;
                case Str_W_AsgRSh:  value = *scopestack[scopelevel].varptr >> value; break;
                case Str_W_Inc:     value = *scopestack[scopelevel].varptr +  1;           break;
                case Str_W_Dec:     value = *scopestack[scopelevel].varptr -  1;           break;
                case Str_W_AsgDiv:  if (value == 0) value = 0xffffffff; else
                                    value = *scopestack[scopelevel].varptr /  value; break;
                case Str_W_AsgMod:  if (value != 0)
                                    value = *scopestack[scopelevel].varptr %  value; break;
            }
            /*
            This code is used for clipping values at overflow or
            underflow, rather than wrapping them around.
            i.e. MAXINT + 1 = MAXINT, rather than the usual MAXINT + 1 = 0

            if (value & 0x80000000) value = 0;
            if (value & 0x00010000) value = 0xffff;
            */
            if (skip == 0) {
                *scopestack[scopelevel].varptr = value;
            }
            PopScope (this);
            return;
        }
        
        /* Common arithmetic; the result is pushed back onto the
         * stack; i.e. + - >= << etc
         */
        switch (scopestack[scopelevel].action) {
            case Str_W_Power:   j=1;
                                for (i=0; i<value; i++) j *= scopestack[scopelevel].value;
                                value = j;                                                       action = P_ValueEnd; return;
            case Str_W_Divide:  if (value == 0) value = 0xffffffff; else
                                value = scopestack[scopelevel].value /  value; action = P_ValueEnd; return;
            case Str_W_Modulo:  if (value != 0)
                                value = scopestack[scopelevel].value %  value; action = P_ValueEnd; return;
            case Str_W_Mult:    value = scopestack[scopelevel].value *  value; action = P_ValueEnd; return;
            case Str_W_Add:     value = scopestack[scopelevel].value +  value; action = P_ValueEnd; return;
            case Str_W_Minus:   value = scopestack[scopelevel].value -  value; action = P_ValueEnd; return;
            case Str_W_And:     value = scopestack[scopelevel].value &  value; action = P_ValueEnd; return;
            case Str_W_Xor:     value = scopestack[scopelevel].value ^  value; action = P_ValueEnd; return;
            case Str_W_Or:      value = scopestack[scopelevel].value |  value; action = P_ValueEnd; return;
            case Str_W_LShift:  value = scopestack[scopelevel].value << value; action = P_ValueEnd; return;
            case Str_W_RShift:  value = scopestack[scopelevel].value >> value; action = P_ValueEnd; return;
            case Str_W_LogAnd:  value = scopestack[scopelevel].value && value; action = P_ValueEnd; return;
            case Str_W_LogOr:   value = scopestack[scopelevel].value || value; action = P_ValueEnd; return;
            case Str_W_Equals:  value = scopestack[scopelevel].value == value; action = P_ValueEnd; return;
            case Str_W_InEqual: value = scopestack[scopelevel].value != value; action = P_ValueEnd; return;
            case Str_W_GreatEq: value = scopestack[scopelevel].value >= value; action = P_ValueEnd; return;
            case Str_W_Greater: value = scopestack[scopelevel].value >  value; action = P_ValueEnd; return;
            case Str_W_LessEq:  value = scopestack[scopelevel].value <= value; action = P_ValueEnd; return;
            case Str_W_Less:    value = scopestack[scopelevel].value <  value; action = P_ValueEnd; return;
            case Str_W_Not:     value = !value;                                            action = P_ValueEnd; return;
            case Str_W_Negate:  value = (value ^ -1) + 1;                                  action = P_ValueEnd; return;
            case P_CaseRun:
            case Str_W_BrackL:
                action = P_Command;
                scopelevel++;
                return;

			/* Flow control. For instance, 'if' causes the next block to be
             * skipped if the value on top of the stack is considered false
             */
            case Str_W_If:
                action = P_Then;
                if (skip == 0 && value == 0) skip = Skip_If;
                PushScope (this, P_Command);
                return;
            case Str_W_Switch:
                action = P_SwitchEnd;
                if (skip) {
                    PushScope (this, P_Switch);
                    action = P_SwBreak_0;
                } else {
                    PushScope (this, P_Switch);
                    action = P_Switch_0;
                }
                return;
            case P_SwitchEnd:
                if (skip == Skip_Break) skip = 0;
                PopScope (this);
                return;
            case Str_W_Case:
                PopScope (this);
                return;
            case P_Case:
                if (value == scopestack[scopelevel - 1].value) {
                    action = P_CaseRun;
                    PushScope (this, P_Command);
                } else {
                    action = P_Switch;
                    value = scopestack[scopelevel].value;
                }
                return;
            case Str_W_Else:
                PopScope (this);
                return;
            case Str_W_While:
                action = P_Loop;
                if (skip == Skip_Break
                || (skip == 0 && value == 0)) skip = Skip_While;
                PushScope (this, P_Command);
                return;
            case P_For4:
                pos = scopestack[scopelevel].retpos;
                action = P_For0;
                PushScope (this, P_Command);
                skip = Skip_For1;
                return;
            case P_For0:
                action = P_For1;
                PushScope (this, P_Value);
                return;
            case P_For1:
                action = P_For2;
                if (skip == Skip_Break
                || (skip == 0 && value == 0)) skip = Skip_For;
                PushScope (this, P_CmdVar);
                if (skip == 0) skip = Skip_For1;
                return;
            case P_FuncRef1:
                funcparm[funcparmno] = value;
                funcparmno++;
                action = P_FuncRef1;
                return;
            case P_FuncCall:
                pos    = scopestack[scopelevel].retpos;
                value  = 0;
                action = P_ValueEnd;
                return;
        }
        action = scopestack[scopelevel].action;
        if (skip == Skip_For2) skip = 0;
    }
}

bool MakeByteCode (struct CParser *this) {
    unsigned bcpos, i, n,
             value, theline = 1;
    char     word     [33],
             *spcword [1000];
    memset (bytecode, 0, 0xffff);
    memset (spcword, NULL, sizeof(spcword));
    pos = 0;
    for (i=Str_W_Last + 1, n=0; i<Str_SysLast; i++, n++) if (StrTable[i]) {
        spcword[n] = strdup(StrTable[i]);
    }
    for (bcpos = 0; pos < scrlen; bcpos ++) {
        while (script[pos] <= ' ' || script[pos] > '~' || script[pos] == ';' || script[pos] == ':') {
            if (script[pos] == 0x0a) theline ++;
            pos++;
            if (pos >= scrlen) break;
        }
        if (pos >= scrlen) break;
        
        /* Numerical constants; note that these are taken modulo MAXINT */
        if (script[pos] >= '0' && script[pos] <= '9') {
            value = 0;
            while (script[pos] >= '0' && script[pos] <= '9') {
                value = (value * 10) + script[pos] - '0';
                pos++;
                if (pos >= scrlen) break;
            }
            linenum [bcpos]   = theline;
            bytecode[bcpos++] = Str_W_Const;
            linenum [bcpos]   = theline;
            bytecode[bcpos]   = value;
            
        /* Anything alpabetical. This is first compared to known words (i.e. if,
         * while, for...) and if no match is found, it is considered to be a variable.
         */
        } else if ((script[pos] >= 'a' && script[pos] <= 'z') || script[pos] == '_') {
            word[0] = script[pos];
            for (n=1; 1; n++) {
                if (n >= MAX_IDENT) {
                    PrintError (this, Str_S_LongID);
                    return false;
                }
                if ((script[pos + n] < 'a' || script[pos + n] > 'z')
                &&  (script[pos + n] < '0' || script[pos + n] > '9')
                &&   script[pos + n] != '_') break;
                if (pos + n >= scrlen) break;
                word[n] = script[pos + n];
            }
            word[n] = 0;
            linenum [bcpos] = theline;
            bytecode[bcpos] = P_Unknown;
            pos += n;
            for (i=Str_W_First+1; i<Str_W_Const; i++) {
                if (strcmp (StrTable[i], word) == 0) {
                    bytecode[bcpos] = i;
                    break;
                }
            }
            if (bytecode[bcpos] == P_Unknown) for (i=Str_W_Const+1; i<Str_W_Nop; i++) {
                if (strcmp (StrTable[i], word) == 0) {
                    bytecode[bcpos] = Str_W_Const;
                    value = Constant (this, i);
                    break;
                }
            }
            if (bytecode[bcpos] == P_Unknown) {
                for (i=0; i<1000; i++) {
                    if (spcword[i] == NULL) {
                        spcword[i] = strdup (word);
                        bytecode[bcpos] = Str_W_Last + i + 1;
                        break;
                    }
                    if (strcmp (spcword[i], word) == 0) {
                        bytecode[bcpos] = Str_W_Last + i + 1;
                        break;
                    }
                }
            }
            
        /* Symbols, such as { == and */
        } else {
            for (i=Str_W_Nop; i<Str_W_Last; i++) {
                if (script[pos] == StrTable[i][0]) {
                    if (StrTable[i][1] == script[pos + 1]) {
                        if (StrTable[i][2] == script[pos + 2]) pos++;
                        else if (StrTable[i][2] > 32) continue;
                        pos++;
                    } else if (StrTable[i][1] > 32) continue;
                    pos++;
                    linenum [bcpos] = theline;
                    bytecode[bcpos] = i;
                    break;
                }
            }
        }
        
        /* Strip comments */
        if (bytecode[bcpos] == Str_W_Comment
        ||  bytecode[bcpos] == Str_W_Comm1) {
            while (pos < scrlen && script[pos] >= ' ') pos++;
            bcpos --;
        }
        if (bytecode[bcpos] == Str_W_ComLong) {
            while (pos < scrlen - 1 && (script[pos] != '*' || script[pos + 1] != '/')) {
                if (script[pos] == 0x0a) theline ++;
                pos++;
            }
            pos += 2;
            bcpos --;
        }
        
        /* If tracing, print the entire bytecode */
        if (tracelevel >= 4) {
            if (bytecode[bcpos - 1] == Str_W_Const) {
                printf ("%3.3d %4.4x '%d'\n", theline, 0, bytecode[bcpos]);
            } else if (bytecode[bcpos] > Str_W_Last) {
                printf ("%3.3d %4.4x '%s'\n", theline, bytecode[bcpos],
                        spcword[bytecode[bcpos] - Str_W_Last - 1]);
            } else {
	            printf ("%3.3d %4.4x %s\n", theline, bytecode[bcpos], StrTable[bytecode[bcpos]]);
            }
        }
    }
    if (tracelevel >= 2) {
    	for (i=0; i<1000; i++) if (spcword[i])
            printf ("%3.3d %s\n", i + 1, spcword[i]);
    }
    bytecode[bcpos] = Str_W_Last;
    linenum [bcpos] = theline;
    pos = 0;
    return true;
}

/* Process: Reads the next token and processes it. */
unsigned Process (struct CParser *this) {
    unsigned token   = 0,
             newword = 0,
             i; 
    char     pbuf  [255];
    struct CVar     *temp;
    if (end) return Str_S_None;

    if (pos >= scrlen) {
        token = Str_W_Last;
    } else {
        token 		  = bytecode[pos];
        curline = linenum [pos];
        pos ++;
        if (token == Str_W_Const) {
            value = bytecode[pos];
            pos ++;
        } else if (token > Str_W_Last) {
            newword = token - Str_W_Last;
            token = P_Unknown;
            for (temp = varlist; temp != NULL; temp = temp->next) {
                if (temp->tokenref == newword) {
                    if (temp->is_func) token = P_FuncRef;
                    else               token = P_VarRef;
                    varptr = &temp->value;
                    break;
                }
            }
        }
    }

    switch (token) {
        case Str_W_Quote:   for (i=0; i<sizeof (pbuf) - 1; i++) {
                                if (script[pos] == '\\' && script[pos + 1] == 'n') {
                                    pbuf[i] = 0x0a;
                                    pos += 2;
                                } else {
                                    pbuf[i] = script[pos];
                                    pos ++;
                                }
                                if (pos >= scrlen - 1 || script[pos] < 32 || script[pos] == '"') break;
                            }
                            pbuf[i + 1] = 0;
                            if (i == sizeof (pbuf) - 1) {
                                while (pos < scrlen - 1 && script[pos] != '"' && script[pos] >= ' ')
                                    pos ++;
                            }
                            pos ++;
                            break;
        case Str_W_Equals1: token = Str_W_Equals;  break;
        case 0:             return Str_S_Unknown;
    }
    if (tracelevel >= 3) switch (token) {
        case P_Unknown: Trace (this, 3, StrTable[Str_S_TNew],  newword, " "); break;
        case P_FuncRef: Trace (this, 3, StrTable[Str_S_TFunc], newword, " "); break;
        case P_VarRef:  Trace (this, 3, StrTable[Str_S_TVar],  newword, " "); break;
        case Str_W_Const:Trace (this, 3, StrTable[token],       value, " "); break;
        default:        Trace (this, 3, StrTable[token],       0,       " ");
    }
    temp = (struct CVar*) varptr;
    while (1) switch (action) {
        case Str_W_Main:
            end = 1;
            if (token != Str_W_Last) return Str_S_NoEOF;
            return Str_S_None;

        case P_Start:
            switch (token) {
                case Str_W_Var:  PushScope (this, P_NewVar);  return 0;
                case Str_W_Func: PushScope (this, P_NewFunc); return 0;
                case Str_W_Main: action = Str_W_Main;
                                 if (mainpos	 == 0) skip = Skip_Stack;
                                 mainpos	 = pos;
                                 Trace (this, 1, StrTable[Str_S_Main], 0, " ");
                                 PushScope (this, P_Command); return 0;
                case Str_W_Last: return Str_S_NoMain;
            }
            return Str_S_Start;

        case P_CmdVar:
            switch (token) {
                case P_VarRef:
                case P_FuncRef:  action = token; return 0;
            }
            return Str_S_VarCmd;

        case P_Command:
            switch (token) {
                case Str_W_Var:                       PushScope (this, P_NewVar);   return 0;
                case Str_W_Switch:
                    if (scopestack[scopelevel - 1].action == P_CaseRun) return Str_S_SwNest;
                case Str_W_While:  retpos = pos;
                case Str_W_Case:
                case Str_W_Return:
                case Str_W_Print:
                case Str_W_If:     action = token;    PushScope (this, P_Value);    return 0;
                case Str_W_Do:     retpos = pos;
                case Str_W_BrackL: action = token;    PushScope (this, P_Command);  return 0;
                case P_VarRef:
                case P_FuncRef:
                case Str_W_For:    action = token;                            return 0;
                case Str_W_Default:                                           return 0;
                case Str_W_Func:   return Str_S_Func;
                case Str_W_Quote:  if (skip == 0) printf (pbuf); return 0;
                case Str_W_Endl:   if (skip == 0) printf ("\n"); return 0;
                case Str_W_Cont:
                case Str_W_Break:
                    for (i=scopelevel-1; i>0; i--) switch (scopestack[i].action) {
                        case P_CaseRun:
                            if (token == Str_W_Cont) continue;
                            if (skip == 0) {
                                scopelevel = i - 1;
                                pos        = scopestack[i - 1].retpos;
                                action     = Str_W_Switch;
                                skip       = Skip_Break;
                                PushScope (this, P_Value);
                            }
                            return 0;
                        case P_Loop:
                            if (skip == 0) {
                                scopelevel = i;
                                pos        = scopestack[i].retpos;
                                action     = Str_W_While;
                                if (token == Str_W_Break) skip = Skip_Break;
                                PushScope (this, P_Value);
                            }
                            return 0;
                        case P_For3:
                            if (skip == 0) {
                                scopelevel = i;
                                pos = scopestack[scopelevel].retpos;
                                action = P_For4;
                                if (token == Str_W_Break) skip = Skip_Break;
                                PushScope (this, P_Command);
                                skip = Skip_For2;
                                PushScope (this, P_Value);
                                if (token == Str_W_Break) skip = Skip_Break;
                                PushScope (this, P_Command);
                            }
                            return 0;
                    }
                    if (token == Str_W_Cont) return Str_S_Cont;
                    return Str_S_Break;
                case Str_W_BrackR:
                    if (scopestack[scopelevel - 1].action != Str_W_BrackL
                    &&  scopestack[scopelevel - 1].action != P_CaseRun)   return Str_S_RgtBrac;
                    scopelevel--;
                    PopScope (this);
                    KillVars (this);
                    return 0;
                case Str_W_Last: return Str_S_EOF;
            }
            return Str_S_Cmd;

        case P_SwBreak_0:
        case P_Switch_0:
            if (token != Str_W_BrackL) return Str_S_SwBrac;
            action++;
            return 0;

        case P_SwBreak:
            if (token == Str_W_BrackR) PopScope (this);
            if (token == Str_W_BrackL) {
                PushScope (this, Str_W_BrackL);
                PushScope (this, P_Command);
            }
            return 0;

        case P_Switch:
            switch (token) {
                case Str_W_BrackL:  PushScope (this, Str_W_BrackL);
                                    skip = Skip_While;
                                    PushScope (this, P_Command);
                                    return 0;
                case Str_W_Var:     PushScope (this, P_NewVar);
                                    return 0;
                case Str_W_Case:    action = P_Case;
                                    PushScope (this, P_Value);
                                    return 0;
                case Str_W_BrackR:  PopScope (this);
                                    return 0;
                case Str_W_Switch:  return Str_S_SwNest;
                case Str_W_Default: action = P_CaseRun;
                                    PushScope (this, P_Command);
                                    return 0;
            }
            return 0;

        case Str_W_For:
            if (token != Str_W_ParenL) return Str_S_ForParn;
            action = P_For0;
            retpos = pos;
            PushScope (this, P_Command);
            return 0;

        case P_For2:
            if (token == Str_W_ParenR) {
                action = P_For3;
                PushScope (this, P_Command);
                return 0;
            }
            return Str_S_ForParn;

        case P_For3:
            if (skip == 0) {
                pos = scopestack[scopelevel].retpos;
                action = P_For4;
                PushScope (this, P_Command);
                skip = Skip_For2;
                PushScope (this, P_Value);
                PushScope (this, P_Command);
                return 0;
            }
            if (skip == Skip_For) skip = 0;
            PopScope (this);
            break;

        case Str_W_Do:
            if (token != Str_W_While) return Str_S_While;
            action = P_Do;
            PushScope (this, P_Value);
            return 0;

        case P_Do:
            if (skip == 0 && value != 0) {
                pos = retpos;
                action = Str_W_Do;
                PushScope (this, P_Command);
                return 0;
            }
            PopScope (this);
            break;

        case P_Loop:
            if (skip == 0) {
                pos = scopestack[scopelevel].retpos;
                action = Str_W_While;
                PushScope (this, P_Value);
                return 0;
            }
            if (skip == Skip_While) skip = 0;
            PopScope (this);
            break;

        case P_Then:
            if (token == Str_W_Else) {
                action = Str_W_Else;
                if      (skip == Skip_If) skip = 0;
                else if (skip == 0)       skip = Skip_If;
                PushScope (this, P_Command);
                return 0;
            }
            PopScope (this);
            break;

        case P_VarRef:
            if (temp->arraysize) {
                if (token != Str_W_ArrayL) return Str_S_Array;
                PushScope (this, Str_W_ArrayL);
                PushScope (this, P_Value);
                return 0;
            }
            if (token >= Str_W_Assign && token <= Str_W_Dec) {
                action = token;
                PushScope (this, P_Value);
                if (token == Str_W_Inc || token == Str_W_Dec) PopScope (this);
                return 0;
            }
            return Str_S_ReqAsg;

        case P_Value:
            switch (token) {
                case P_VarRef:     if (temp->arraysize == 0) {
                                       value = *varptr; break;
                                   }
                                   action = P_ArrayRef;
                                   return 0;
                case P_FuncRef:    action = token; return 0;
                case Str_W_Const:  break;
                case Str_W_ParenL:
                case Str_W_Not:    action = token;        PushScope (this, P_Value); return 0;
                case Str_W_Minus:  action = Str_W_Negate; PushScope (this, P_Value); return 0;
                default:           return Str_S_Expr;
            }
            action = P_ValueEnd;
            return 0;

        case P_ValueEnd:
            if (token >= Str_W_LogAnd && token <= Str_W_Or) {
                i = scopestack[scopelevel - 1].action;
                if (i >= Str_W_LogAnd && i <= Str_W_Or && i > token) PopScope (this);
                action = token;
                PushScope (this, P_Value);
                return 0;
            } else {
                PopScope (this);
                if (token == Str_W_ParenR) {
                    while (action == P_ValueEnd) PopScope (this);
                    if (action == P_FuncRef1) break;
                    if (skip == Skip_For1) return 0;
                    if (action == P_For2) {
                        action = P_For3;
                        PushScope (this, P_Command);
                    } else {
                        if (action != Str_W_ParenL) return Str_S_RgtParn;
                        action = P_ValueEnd;
                    }
                    return 0;
                } else {
                    if (action == P_For2) return Str_S_ForParn;
                    if (action == Str_W_ParenL) return Str_S_RgtParn;
                }
                break;
            }

        case Str_W_Print:
            if (value & 0x80000000) value = 0;
            if (value & 0x00010000) value = 0xffff;
            if (skip == 0) {
                printf (" %6d ", value);
            }
            action = P_Command;
            break;

        case P_NewVar:
            if (token != P_Unknown) {
                if (temp == NULL)                        return Str_S_DupDef;
                if (temp->is_func)                       return Str_S_DupDef;
                if (temp->scope >= scopelevel - 1) return Str_S_DupDef;
            }
            AddVar (this, newword, P_NewVar);
            action++;
            return 0;

        case P_NewVarEnd:
            switch (token) {
                case Str_W_Comma:  action = P_NewVar; return 0;
                case Str_W_Assign: PushScope (this, Str_W_Assign);
                                   varptr = scopestack[scopelevel - 1].varptr;
                                   PushScope (this, P_Value); return 0;
                case Str_W_ArrayL: action = P_NewArrayEnd;
                                   PushScope (this, P_Value); return 0;
            }
            PopScope (this);
            break;

        case P_NewArrayEnd:
            if (token != Str_W_ArrayR) return Str_S_RgtBrck;
            if (value < 2 || value > 4096) return Str_S_ArBound;
            temp->arraysize = value;
            temp->arraydata = malloc (value * sizeof (unsigned));
            if (temp->arraydata == NULL) {
                printf ("Memory error\n");
                exit (1);
            }
            memset (temp->arraydata, 0, value * sizeof (unsigned));
            action = P_NewArrayEnd1;
            return 0;

        case P_NewArrayEnd1:
            if (token == Str_W_Comma) {
                action = P_NewVar; return 0;
            }
            PopScope (this);
            break;

        case Str_W_ArrayL:
            if (token != Str_W_ArrayR) return Str_S_RgtBrck;
            if (skip == 0) {
                temp = (struct CVar*) scopestack[scopelevel].varptr;
                if (value >= temp->arraysize) return Str_S_ArrayOv;
                varptr = temp->arraydata + value;
            }
            PopScope (this);
            action = P_ArrayAsg;
            return 0;

        case P_ArrayAsg:
            if (token >= Str_W_Assign && token <= Str_W_Dec) {
                action = token;
                PushScope (this, P_Value);
                if (token == Str_W_Inc || token == Str_W_Dec) PopScope (this);
                return 0;
            }
            return Str_S_ReqAsg;

        case P_ArrayRef:
            if (token != Str_W_ArrayL) return Str_S_Array;
            PushScope (this, P_ArrayRef1);
            PushScope (this, P_Value);
            return 0;

        case P_ArrayRef1:
            if (token != Str_W_ArrayR)    return Str_S_RgtBrck;
            if (skip == 0) {
                temp = (struct CVar*) scopestack[scopelevel].varptr;
                if (value >= temp->arraysize) return Str_S_ArrayOv;
                value = temp->arraydata [value];
            }
            PopScope (this);
            action = P_ValueEnd;
            return 0;

        case P_NewFunc:
            if (token != P_Unknown) return Str_S_DupDef;
            AddVar (this, newword, P_NewFunc);
            action = P_NewFuncEnd;
            return 0;

        case P_NewFuncEnd:
            switch (token) {
                case Str_W_ParenL: action = P_FuncParm; return 0;
                case Str_W_BrackL: action = Str_W_BrackL;
                                   skip   = Skip_Func;
                                   PushScope (this, P_Command);
                                   return 0;
            }
            return Str_S_FuncDef;

        case P_FuncParm:
            switch (token) {
                case Str_W_Comma:   return 0;
                case Str_W_ParenR:  action = P_NewFuncEnd; return 0;
                case P_Unknown:     temp->value++;
                                    if (temp->value >= MAX_PARM) return Str_S_ParmOv;
                                    scopelevel++;
                                    AddVar (this, newword, P_NewVar);
                                    scopelevel--;
                                    varptr = (unsigned*) temp;
                                    return 0;
            }
            return Str_S_DupDef;

        case P_FuncRef:
            if (token != Str_W_ParenL) return Str_S_FuncCl;
            action     = P_FuncRef1;
            funcptr    = temp;
            funcparmno = 0;
            return 0;

        case P_FuncRef1:
            switch (token) {
                case Str_W_Comma:   return 0;
                case Str_W_ParenR:  if (funcparmno > funcptr->value) return Str_S_FuncPrM;
                                    if (funcparmno < funcptr->value) return Str_S_FuncPrF;
                                    if (skip) {
                                        value = 0;
                                        action = P_ValueEnd;
                                        return 0;
                                    }
                                    if (funcptr->arraysize) {
                                        value  = SysCall (this, funcptr->arraysize);
                                        action = P_ValueEnd;
                                        return 0;
                                    }
                                    action = P_FuncEntry;
                                    retpos = pos;
                                    value  = 0;
                                    pos    = funcptr->scope;
                                    return 0;
                case P_FuncRef:     return Str_S_FuncNst;
            }
            PushScope (this, P_Value);
            value = scopestack[scopelevel - 1].value;
            break;

        case P_FuncEntry:
            if (token != Str_W_ParenL) return Str_S_FuncDef;
            action = P_FuncEnt1;
            return 0;

        case P_FuncEnt1:
            switch (token) {
                case P_VarRef:
                case P_Unknown:    scopelevel+=2;
                                   AddVar (this, newword, P_NewVar);
                                   ((struct CVar*) varptr)->value = funcparm[value];
                                   value++;
                                   scopelevel-=2;
                                   return 0;
                case Str_W_Comma:  return 0;
                case Str_W_ParenR: action = P_FuncCall;
                                   PushScope (this, P_Command);
                                   return 0;
                case P_FuncRef:    return Str_S_VarFunc;
            }
            return Str_S_FuncDef;

        case Str_W_Return:
            if (skip == 0) {
                for (i=scopelevel-1; i>0; i--) {
                    if (scopestack[i].action == P_FuncCall) {
                        scopelevel = i;
                        pos = scopestack[i].retpos;
                        KillVars (this);
                        action = P_ValueEnd;
                        return 0;
                    }
                }
                end = 1;
                return 0;
            }
            action = P_ValueEnd;
            break;

        default:
            if (action <= P_Unknown) {
                printf ("action = %d\n", action);
            } else PrintError (this, action);
            break;
    }

    return Str_ErrLast;
}
