// overkill on the debugmsgs (and make some of them important)

#include "Common/StdAfx.h"
#include "Kernel.h"
#include "Socket.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


#ifdef LINUX

//#warning fixme
// cut/paste to stdafx

typedef int WSADATA;
static void WSACleanup  (void) {}
static bool WSAStartup  (WORD, WSADATA*) { return false; }
static void closesocket (SOCKET) {}
static bool ioctlsocket (SOCKET, int, unsigned long*) { return false; }

enum {
    FIONBIO         = 1,
    SOCKET_ERROR    = 2,
    WSAECONNRESET   = 3,
    WSAECONNABORTED = 4,
    WSABASEERR      = 5
};

#endif


CPSocket Socket;

CPSocket::CPSocket (void) {
    memset (this, 0, sizeof (CPSocket));
    port  = DEFAULT_PORT;
}

CPSocket::~CPSocket (void) {
    DeInit     ();
    WSACleanup ();
}

void CPSocket::Init (void) {
    WORD    ver = 0x0101;
    WSADATA dat;
    if (theApp.network == false) return;
    state = SK_Disabled;
    if (WSAStartup (ver, &dat)) {
        theApp.DebugMessage ("Winsock failed to initialize.");
        theApp.network = false;
    } else if (gethostname (hostname, sizeof (hostname)) != 0) {
        theApp.DebugMessage ("Winsock: Gethostname failed.");
        theApp.network = false;
    } else {
        theApp.DebugMessage ("Winsock initialized.");
        state = SK_Inactive;
    }
}

bool CPSocket::GetHost (const char *name) {
    hostent       *hostptr;
    unsigned long hostaddr;
    char          buf[255];

    CSocket s1;
    CAsyncSocket nn;

    if (port < 1024 || port > 65535) return false;
    hostptr = gethostbyname (name);
    if (hostptr == NULL) {
        theApp.DebugMessage ("Gethostbyname failed.");
        return false;
    }
    hostaddr = *((unsigned long*) hostptr->h_addr);
    if (state != SK_ClientBroken) {
        sprintf (buf, "Address = %ld.%ld.%ld.%ld", hostaddr & 0xff, (hostaddr >> 8) & 0xff, (hostaddr >> 16) & 0xff, (hostaddr >> 24) & 0xff);
        theApp.DebugMessage (buf);
    }
    addr.sin_family      = AF_INET;
    addr.sin_port        = htons (port);
    addr.sin_addr.s_addr = hostaddr;
    svr                  = socket (PF_INET, SOCK_STREAM, 0);
    // may want to try SOCK_RDM as an alternative

    if (svr <= 0) {
        theApp.DebugMessage ("Socket init failed.");
        return false;
    }
    return true;
}

bool CPSocket::InitServer (void) {
    unsigned long n;
    switch (state) {
        case SK_Disabled: return false;
        case SK_Server:   return true;
        case SK_Client:   DeInit ();
        default:          break;
    }
    if (GetHost (hostname) == false) return false;
    n = bind (svr, (sockaddr*) &addr, sizeof (addr));
    if (n != 0) {
        theApp.DebugMessage ("Socket bind failed.");
        return false;
    }
    SetAsync (svr);
    n = listen (svr, 5);
    if (n != 0) {
        theApp.DebugMessage ("Socket listen failed.");
        return false;
    }
    theApp.DebugMessage ("Server on line.");
    state = SK_Server;
    return true;
}

bool CPSocket::InitClient (void) {
    switch (state) {
        case SK_Disabled: return false;
        case SK_Client:
        case SK_Server:   DeInit ();
        default:          break;
    }
    state = SK_ClientBroken;
    return GetHost (usename);
}

void CPSocket::DeInit (void) {
    unsigned i;
    switch (state) {
        case SK_Server:
            for (i=0; i<MAX_LINK; i++) if (sock[i] != 0) {
                closesocket (sock[i]);
                sock[i] = 0;
            }
            closesocket (svr);
            state = SK_Inactive;
            theApp.DebugMessage ("Server now off line.");
            break;
        case SK_ClientBroken:
        case SK_Client:
            closesocket (svr);
            state = SK_Inactive;
            theApp.DebugMessage ("Client now off line.");
            break;
        default:
            break;
    }
}

void CPSocket::SetAsync (unsigned sock) {
    unsigned long n = 1;
    if (ioctlsocket (sock, FIONBIO, &n) != 0)
        theApp.DebugMessage ("Socket ioctl failed.");
}

void CPSocket::SetLink (unsigned nlink) {
    if (nlink >= MAX_LINK) link = 0;
    else link = nlink;
}

void CPSocket::Unlink (void) {
    recvpos[link] = 0xffffffff;
}

void CPSocket::SetOwner (unsigned p, bool first) {
    if (first) owner[link * 2]     = p;
    else       owner[link * 2 + 1] = p;
    char buf[255];
    sprintf (buf, "Socket %d owner %d = %d", link, first, p);
    theApp.DebugMessage (buf);
}

bool CPSocket::CheckOwner (unsigned p) {
    if (owner[link * 2]     == p) return true;
    if (owner[link * 2 + 1] == p) return true;
    return false;
}

void CPSocket::SendByte (char byte) {
    if (sendpos < MAX_BUFFER) {
        sendbuf[sendpos] = byte;
        sendpos++;
        //theApp.DebugMessage (MkStr (byte));
    }
}

void CPSocket::SendWord (short word) {
    if (sendpos < MAX_BUFFER - 1) {
        sendbuf[sendpos    ] =  word       & 0xff;
        sendbuf[sendpos + 1] = (word >> 8) & 0xff;
        sendpos += 2;
    }
}

void CPSocket::SendDW (unsigned dw) {
    if (sendpos < MAX_BUFFER - 3) {
        sendbuf[sendpos    ] =  dw        & 0xff;
        sendbuf[sendpos + 1] = (dw >>  8) & 0xff;
        sendbuf[sendpos + 2] = (dw >> 16) & 0xff;
        sendbuf[sendpos + 3] = (dw >> 24) & 0xff;
        sendpos += 4;
    }
}

void CPSocket::Send (const void *ptr, unsigned len) {
    unsigned i;
    for (i=0; i<len; i++) {
        if (sendpos >= MAX_BUFFER) break;
        sendbuf[sendpos] = ((char*) ptr) [i];
        sendpos++;
    }
}

char CPSocket::RecvByte (void) {
    unsigned l;
    char     n;
    if      (state == SK_Client) l = 0;
    else if (state == SK_Server) l = link;
    else return 0;
    if (sock[l] == 0) return 0;
    if (recvpos[l] >= recvend[l]) return 0;
    n = recvbuf[l][recvpos[l]];
    recvpos[l]++;
    return n;
}

short CPSocket::RecvWord (void) {
    unsigned l;
    short    n;
    if      (state == SK_Client) l = 0;
    else if (state == SK_Server) l = link;
    else return 0;
    if (sock[l] == 0) return 0;
    if (recvpos[l] >= recvend[l]) return 0;
    n = recvbuf[l][recvpos[l]] + (recvbuf[l][recvpos[l] + 1] << 8);
    recvpos[l] += 2;
    return n;
}

unsigned CPSocket::RecvDW (void) {
    unsigned l;
    short    n;
    if      (state == SK_Client) l = 0;
    else if (state == SK_Server) l = link;
    else return 0;
    if (sock[l] == 0) return 0;
    if (recvpos[l] >= recvend[l]) return 0;
    n =  recvbuf[l][recvpos[l]]
      + (recvbuf[l][recvpos[l] + 1] <<  8)
      + (recvbuf[l][recvpos[l] + 2] << 16)
      + (recvbuf[l][recvpos[l] + 3] << 24);
    recvpos[l] += 4;
    return n;
}

void CPSocket::Recv (void *buf, unsigned len) {
    unsigned l, i;
    for (i=0; i<len; i++) ((char*) buf) [i] = 0;
    if      (state == SK_Client) l = 0;
    else if (state == SK_Server) l = link;
    else return;
    if (sock[l] == 0) return;
    for (i=0; i<len; i++) {
        if (recvpos[l] >= recvend[l]) return;
        ((char*) buf) [i] = recvbuf[l][recvpos[l]];
        recvpos[l]++;
    }
}

void CPSocket::RecvSkip (unsigned len) {
    unsigned l;
    if      (state == SK_Client) l = 0;
    else if (state == SK_Server) l = link;
    else return;
    recvpos[l] += len;
    if (recvpos[l] > recvend[l]) recvpos[l] = recvend[l];
}

void CPSocket::Update (void) {
    char     buf[1024], b1[255];
    int i, j;
    if (theApp.network == false) return;
    if (state == SK_ClientBroken) {
        /*if (GetHost (usename) == false) {
            state = SK_Inactive;
            return;
        }*/
        if (connect (svr, (sockaddr*) &addr, sizeof (addr)) != 0) return;
        SetAsync (svr);
        theApp.DebugMessage ("Client on line.");
        state = SK_Client;
        return;
    }
    if (state != SK_Client && state != SK_Server) return;
    for (link=0; (state == SK_Server ? link<MAX_LINK : link < 1); link++) {
        if (recvpos[link] == 0xffffffff) {
            sprintf (buf, "Shutting off link %d", link);
            theApp.DebugMessage (buf);
            buf[0] = -1;
            i = send (sock[link], buf, 10, 0);
            theApp.DebugMessage (MkStr (i));
            closesocket (sock[link]);
            sock[link]    = 0;
            recvpos[link] = 0;
            continue;
        }
        if (state == SK_Client) {
            i = recv (svr, buf, sizeof (buf), 0);
            sprintf (b1, "Recv on server link / port %d: %d bytes", svr, i);
            theApp.DebugMessage (b1);
        } else {
            if (sock [link] == 0) continue;
            i = recv (sock [link], buf, sizeof (buf), 0);
            sprintf (b1, "Recv on link %d / port %d: %d bytes", link, sock[link], i);
            theApp.DebugMessage (b1);
        }
        if (i != -1 && i != 0) {
            if (i == WSAECONNRESET || i == WSAECONNABORTED) {
                sprintf (buf, "Link %d has been cut.", link);
                theApp.DebugMessage (buf);
                if (state == SK_Client) {
                    closesocket (svr);
                    state = SK_ClientBroken;
                } else {
                    closesocket (sock[link]);
                    sock[link] = 0;
                }
                continue;
            } else if (i >= WSABASEERR) {
                sprintf (buf, "Link %d gives message %d", link, i - WSABASEERR);
                theApp.DebugMessage (buf);
            } else {
                if (recvpos[link] != 0) {
                    for (j=0; recvpos[link] + j < recvend[link]; j++)
                        recvbuf[j][link] = recvbuf[link][recvpos[link] + j];
                    recvend[link] -= recvpos[link];
                    recvpos[link] = 0;
                }
                for (j=0; j < i; j++, recvend[link]++)
                    recvbuf[link][recvend[link]] = buf[j];
            }
        }
        if (sendpos == 0) continue;
        if (state == SK_Client) {
            i = send (svr, sendbuf, sendpos, 0);
            sprintf (buf, "Send to server / port %d: %d bytes of %d", svr, i, sendpos);
            theApp.DebugMessage (buf);
        } else {
            i = send (sock[link], sendbuf, sendpos, 0);
            sprintf (buf, "Send to link %d / port %d: %d bytes of %d", link, sock[link], i, sendpos);
            theApp.DebugMessage (buf);
        }
        if (i == -1 || i == 0 || i == WSAECONNRESET || i == WSAECONNABORTED) {
            sprintf (buf, "Link %d has been cut.", link);
            theApp.DebugMessage (buf);
            if (state == SK_Client) {
                closesocket (svr);
                state = SK_ClientBroken;
            } else {
                closesocket (sock[link]);
                sock[link] = 0;
            }
        } else if (i >= WSABASEERR) {
            sprintf (buf, "Link %d gives message %d\n", link, i - WSABASEERR);
            theApp.DebugMessage (buf);
        }
    }
    sendpos    = 0;
    sendbuf[0] = 0;
    if (state == SK_Server) {
        for (i=0; i<MAX_LINK; i++) if (sock[i] == 0) break;
        if (i != MAX_LINK) {
            sock[i] = accept (svr, (sockaddr*) &addr, NULL);
            if (sock[i] == SOCKET_ERROR) {
                sock[i] = 0;
            } else {
                sprintf (buf, "Client %d has connected at port %d.", i, sock[i]);
                theApp.DebugMessage (buf);
                SetAsync (sock[i]);
            }
        }
    }
}
