#include "Common/StdAfx.h"
#include "Graphics.h"
#include "Kernel.h"
#include "ViewBase.h"
#include "../View.h"
#include "../Strings.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#ifdef SERVERMODE
#define server_return return
#else
#define server_return
#endif



/****************************** Viewer *******************************/

CViewer::CViewer (void) {
    memset (this, 0, sizeof (CViewer));
}

CViewer::~CViewer (void) {
    CViewer *temp;
    while (next != NULL) {
        temp = next->next;
        next->next = NULL;
        delete next;
        next = temp;
    }
}

CViewer::CViewer (unsigned xp,  unsigned yp,  unsigned xs, unsigned ys,
                  unsigned txp, unsigned typ, unsigned n,  unsigned st) {
    memset (this, 0, sizeof (CViewer));
    xpos      = xp;
    ypos      = yp;
    xsize     = xs;
    ysize     = ys;
    tilexpos  = txp;
    tileypos  = typ;
    num       = n;
    state     = st;
}

void CViewer::Link (CViewer *view) {
    CViewer *temp;
    if (view == NULL) {
        theApp.runlevel = Run_MemError;
    } else {
        for (temp = this; temp->next; temp = temp->next);
        temp->next = view;
    }
}

void CViewer::Move (int x, int y, bool rel) {
    if (rel) {
        if (xpos + x < 0) x = 0 - xpos;
        if (ypos + y < 0) y = 0 - ypos;
        if (next) next->Move (x, y, true);
        xpos += x;
        ypos += y;
    } else {
        if (x == -1) x = xpos;
        if (next) next->Move (x - xpos, y - ypos, true);
        xpos = x;
        ypos = y;
    }
}

bool CViewer::MouseMove (void) {
    if (Screen.mousex >= xpos && Screen.mousex < xpos + xsize
    &&  Screen.mousey >= ypos && Screen.mousey < ypos + ysize) {
        if (state == VS_Disabled) state = VS_MouseDis;
        if (state == VS_Visible)  state = VS_MouseOver;
        if (state != VS_Hidden) {
            P->mouseover = this;
            return true;
        }
    }
    return false;
}

void CViewer::DrawBorder (void) {
    if (xsize && ysize) {
        if (P->mouseover == this)
             Screen.PutEdge (xpos, ypos, xpos + xsize - 1, ypos + ysize - 1, 0xff0000);
        else Screen.PutEdge (xpos, ypos, xpos + xsize - 1, ypos + ysize - 1, 0xffffff);
    }
}

void CViewer::DrawMouse (void) {
    unsigned x, y, my, ny;
    x  = Screen.mousex;
    y  = Screen.mousey;
    my = ypos;
    ny = ypos + ysize;
    if (P->toppanel->link) {
        if (P->toppanel->link->ypos < my) my = P->toppanel->link->ypos;
        if (P->toppanel->link->ypos + P->toppanel->link->ysize > ny)
            ny = P->toppanel->link->ypos + P->toppanel->link->ysize;
    }
    if (x < xpos + 10) x = xpos; else x -= 10;
    if (y < my   + 10) y = my;   else y -= 10;
    if (x > xpos + xsize - 21) x = xpos + xsize - 21;
    if (y > ny           - 21) y = ny           - 21;
    my = Screen.mousecursory + Screen.mousecursorsz * ((Screen.dispstate >> 2) & 7);
    Screen.font->BlitTr (x, y, Screen.mousecursorx, my, Screen.mousecursorsz, Screen.mousecursorsz);
}

void CViewer::DrawBack (unsigned type) {
    unsigned i, j, x, y, sizex, sizey, d = 0;
    if (type == Bk_Parch) {
        if (ysize == 160) {
            Screen.font->BlitTr (xpos,       ypos,       578,      76,      32, 32);
            Screen.font->BlitTr (xpos + 128, ypos,       578 + 64, 76,      32, 32);
            Screen.font->BlitTr (xpos,       ypos + 128, 578,      76 + 64, 32, 32);
            Screen.font->BlitTr (xpos + 128, ypos + 128, 578 + 64, 76 + 64, 32, 32);
            for (i=32; i<128; i+=32) {
                Screen.font->BlitTr (xpos + i,   ypos,       578 + 32, 76,      32, 64);
                Screen.font->BlitTr (xpos + i,   ypos + 96,  578 + 32, 76 + 32, 32, 64);
                Screen.font->BlitTr (xpos,       ypos + i,   578,      76 + 32, 32, 32);
                Screen.font->BlitTr (xpos + 128, ypos + i,   578 + 64, 76 + 32, 32, 32);
                Screen.font->BlitTr (xpos + i,   ypos + 64,  578 + 32, 76 + 32, 32, 32);
            }
        } else {
            Screen.font->BlitTr (xpos,       ypos,       578,      76,      32, 32);
            Screen.font->BlitTr (xpos + 128, ypos,       578 + 64, 76,      32, 32);
            Screen.font->BlitTr (xpos,       ypos +  96, 578,      76 + 64, 32, 32);
            Screen.font->BlitTr (xpos + 128, ypos +  96, 578 + 64, 76 + 64, 32, 32);
            for (i=32; i<128; i+=32) {
                Screen.font->BlitTr (xpos + i,   ypos,       578 + 32, 76,      32, 64);
                Screen.font->BlitTr (xpos + i,   ypos + 64,  578 + 32, 76 + 32, 32, 64);
                if (i < 96) {
                    Screen.font->BlitTr (xpos,       ypos + i,   578,      76 + 32, 32, 32);
                    Screen.font->BlitTr (xpos + 128, ypos + i,   578 + 64, 76 + 32, 32, 32);
                }
            }
        }
    } else {
        switch (type) {
            case Bk_Black:  Screen.PutRect (xpos, ypos, xpos + xsize, ypos + ysize, 0); return;
            case Bk_GrayNE: Screen.Lock    (Screen.back);
                            Screen.PutGray (xpos, ypos, xpos + xsize, ypos + ysize, 0);
                            Screen.Unlock  ();
                            return;
            case Bk_Gray:   Screen.Lock    (Screen.back);
                            Screen.PutGray (xpos, ypos, xpos + xsize, ypos + ysize, 0);
                            Screen.PutLine (xpos,         ypos, xpos + xsize, ypos,         0xffffffff);
                            Screen.PutLine (xpos,         ypos, xpos,         ypos + ysize, 0xffffffff);
                            Screen.PutLine (xpos + xsize, ypos, xpos + xsize, ypos + ysize, 0xbbbbbbbb);
                            Screen.PutLine (xpos, ypos + ysize, xpos + xsize, ypos + ysize, 0xbbbbbbbb);
                            Screen.Unlock  ();
                            return;
            case Bk_Select: sizex = 128; sizey =  40; x = 492; y = 212; d = Screen.animstate & 127; break;
            case Bk_Blue:   sizex = 108; sizey = 108; x = 470; y =  76; break;
            case Bk_Brown:  sizex = 150; sizey = 150; x = 320; y =  75; break;
            case Bk_White:  sizex = 100; sizey = 100; x =   0; y = 210; break;
            case Bk_Water:  sizex =  86; sizey =  86; x = 100; y = 210; break;
            default:        sizex =   0; sizey =   0; x =   0; y =   0;
        }
        if (sizex) {
            if (d) {
                i = d;
                if (i >= xsize) i = xsize;
                for (j = ypos; j + sizey <= ypos + ysize; j += sizey)
                    Screen.font->BlitTr (xpos, j, x + 128 - d, y, i, sizey);
                Screen.font->BlitTr (xpos, j, x + 128 - d, y, i, ypos + ysize - j);
            }
            for (i = xpos + d; i + sizex <= xpos + xsize; i += sizex) {
                for (j = ypos; j + sizey <= ypos + ysize; j += sizey)
                    Screen.font->Blit (i, j, x, y, sizex, sizey);
                Screen.font->Blit (i, j, x, y, sizex, ypos + ysize - j);
            }
            for (j = ypos; j + sizey <= ypos + ysize; j += sizey)
                Screen.font->Blit (i, j, x, y, xpos + xsize - i, sizey);
            if (xpos + xsize > i && ypos + ysize > j)
                Screen.font->Blit (i, j, x, y, xpos + xsize - i, ypos + ysize - j);
        }
    }
}

void CViewer::Press (void) {
    if (P->pressed) {
        if (P->pressed->state == VS_Pressed) {
            if (P->pressed == P->mouseover)
                 P->pressed->state = VS_MouseOver;
            else P->pressed->state = VS_Visible;
        }
    }
    P->pressed    = this;
    P->pressdelay = 20;
    state         = VS_Pressed;
}

void CViewer::PullDown (unsigned *var, unsigned base, unsigned len, unsigned flag) {
    unsigned i, j;
    if (P->pd_var == var) {
        P->pd_x   = 0;
        P->pd_var = NULL;
        return;
    }
    P->pd_x    = xpos;
    P->pd_y    = ypos + ysize;
    P->pd_dx   = xsize;
    P->pd_base = base;
    P->pd_len  = len;
    P->pd_var  = var;
    P->pd_flag = flag;
    if (flag & Txt_AltFont) P->pd_dy = 16;
    else                    P->pd_dy = 22;
    for (i=0; i<len; i++) {
        j = Screen.TextWidth (StrTable[base + i], (flag & Txt_AltFont) > 0) + 4;
        if (j > P->pd_dx) P->pd_dx = j;
    }
    if (xpos + P->pd_dx >= Screen.winxsize)
        P->pd_x = Screen.winxsize - xpos - 1;
    if (ypos + ysize + P->pd_dy * len >= Screen.winysize)
        P->pd_y = ypos - P->pd_dy * len - 4;
}

void CViewer::Event (unsigned event, unsigned) {
    if (event == VE_Draw) {
        switch (state) {
            case VS_Hidden:    break;
            case VS_MouseDis:
            case VS_Disabled:  Screen.font->BlitTr (xpos, ypos, tilexpos, tileypos,             xsize, ysize); break;
            case VS_MouseOver: Screen.font->BlitTr (xpos, ypos, tilexpos, tileypos + ysize,     xsize, ysize); break;
            case VS_Pressed:   Screen.font->BlitTr (xpos, ypos, tilexpos, tileypos + 2 * ysize, xsize, ysize); break;
            default:           Screen.font->BlitTr (xpos, ypos, tilexpos, tileypos,             xsize, ysize); break;
        }
    }
}



/****************************** BackView ******************************/

CBackView::CBackView (int x, int y, unsigned xs, unsigned ys, unsigned type):
    CViewer (x, y, xs, ys, type, 0, 0, VS_Hidden) {
}

void CBackView::Event (unsigned event, unsigned) {
    if (event == VE_Draw) DrawBack (tilexpos);
}



/***************************** PanelBase *****************************/

CPanelBase::CPanelBase (void) {
    memset (this, 0, sizeof (CPanelBase));
}

void CPanelBase::Previous (void) {
    if (panelcount > 0) panelcount--;
    if (lastpanel [panelcount] != NULL && panelcount) {
        lastpanel [panelcount]->Event (VE_DisplayFade);
    }
}

bool CPanelBase::DoPullDown (void) {
    unsigned i;
    if (pd_x) {
        if (Screen.mousex >= pd_x && Screen.mousex < pd_x + pd_dx
        &&  Screen.mousey >= pd_y && Screen.mousey < pd_y + pd_dy * pd_len + 4) {
            i = Screen.mousey - pd_y;
            if (i > 2) i -= 2;
            i /= pd_dy;
            if (i >= pd_len) i = pd_len - 1;
            *pd_var = i;
            pd_x   = 0;
            pd_var = NULL;
            if (P->toppanel) P->toppanel->MouseMove ();
            return true;
        } else {
            pd_x = 0;
            if (P->toppanel) P->toppanel->MouseMove ();
        }
    }
    return false;
}



/***************************** PanelView *****************************/

CPanelView::CPanelView (unsigned type):
    CViewer (0, 0, 0, 0, 0, 0, 0, VS_Visible) {
    if (type != 0xffff) Init (type);
}

void CPanelView::MouseMove (void) {
    CViewer *temp;
    if (P->mouseover != NULL) {
        if (P->mouseover->state == VS_MouseOver) P->mouseover->state = VS_Visible;
        if (P->mouseover->state == VS_MouseDis)  P->mouseover->state = VS_Disabled;
        P->mouseover = NULL;
    }
    if (P->pd_x) {
        if (Screen.mousex >= P->pd_x && Screen.mousex < P->pd_x + P->pd_dx
        &&  Screen.mousey >= P->pd_y && Screen.mousey < P->pd_y + P->pd_dy * P->pd_len + 4) return;
    }
    for (temp = next; temp != NULL; temp = temp->next)
        if (temp->MouseMove ()) break;
    if (temp == NULL && P->toppanel != NULL) {
        if (P->toppanel->link != NULL)
            for (temp = P->toppanel->link->next; temp != NULL; temp = temp->next)
                if (temp->MouseMove ()) break;
    }
}

void CPanelView::Event (unsigned event, unsigned keys) {
    unsigned i, j;
    CViewer  *temp;
    static bool borders = false,
                domouse = false;
    switch (event) {
        case VE_DisplayFadeNoCopy:
        case VE_DisplayFade:
        case VE_Display:
            if (state == VS_Disabled) break;
            if (ysize < Screen.winysize) {
                Screen.Lock (Screen.back);
                if      (P == sc_keyb)  Screen.PutGray (Screen.winxsize / 2 + 8, 0, Screen.winxsize, Screen.winysize, 0);
                else if (P == sc_mouse) Screen.PutGray (0, 0, Screen.winxsize / 2 - 8, Screen.winysize, 0);
                else                    Screen.PutGray (0, 0, Screen.winxsize, Screen.winysize, 0);
                Screen.Unlock ();
                if (event != VE_DisplayFadeNoCopy) Screen.FlipCopy ();
            }
            if (event == VE_Display) P->fade = 20;
            else                     P->fade = 1;
            if (P->lastpanel [P->panelcount] != this) {
                P->lastpanel [P->panelcount] = P->toppanel;
                P->panelcount++;
                if (P->panelcount >= PANEL_MAX) {
                    theApp.runlevel = Run_MemError;
                    return;
                }
            }
            if (P->mouseover) {
                if (P->mouseover->state == VS_MouseOver) P->mouseover->state = VS_Visible;
                if (P->mouseover->state == VS_MouseDis)  P->mouseover->state = VS_Disabled;
                P->mouseover = NULL;
            }
            P->mousehelp = 0;
            P->buffer[0] = 0;
            P->toppanel  = this;
            event        = VE_Update;
            domouse      = true;
            server_return;
            break;
        case VE_LeftClick:
        case VE_RightClick:
            if (event == VE_RightClick) {
                i = RightClick ();
                if (i != 0) {
                    event = i;
                    break;
                }
            }
            i = ypos;
            j = ypos + ysize;
            if (link) {
                if (link->ypos < i)               i = link->ypos;
                if (link->ypos + link->ysize > j) j = link->ypos + link->ysize;
            }
            if (P->mouseover == NULL
            && (Screen.mousex < xpos || Screen.mousex >= xpos + xsize
            ||  Screen.mousey < i    || Screen.mousey >= j)) {
                event = VK_ESCAPE;
                break;
            }
            return;
        case VE_Update:
            server_return;
            break;
        case VE_Draw:
            server_return;
            DrawBack (tilexpos);
            break;
        case VK_APOSTROPHE:
            borders = !borders;
            break;
        default:
            if (P->pd_x) {
                P->pd_x = 0;
                if (P->toppanel) P->toppanel->MouseMove ();
            }
    }
    temp = next;
    i    = 0;
    while (temp != NULL && P->toppanel == this) {
        temp->Event (event, keys);
        temp = temp->next;
        if (temp == NULL && link != NULL && i == 0 && event != VK_APOSTROPHE) {
            temp = link;
            i = 1;
        }
    }
    if (event == VE_Draw && P->pd_x) {
        Screen.PutRect (P->pd_x, P->pd_y, P->pd_x + P->pd_dx, P->pd_y + P->pd_dy * P->pd_len + 4, 0);
        Screen.Lock    (Screen.back);
        Screen.PutEdge (P->pd_x, P->pd_y, P->pd_x + P->pd_dx, P->pd_y + P->pd_dy * P->pd_len + 4, 0xffffff);
        Screen.Unlock  ();
        if (Screen.mousex >= P->pd_x && Screen.mousex < P->pd_x + P->pd_dx
        &&  Screen.mousey >= P->pd_y && Screen.mousey < P->pd_y + P->pd_dy * P->pd_len) {
            i = Screen.mousey - P->pd_y;
            if (i > 2) i -= 2;
            i /= P->pd_dy;
            if (i >= P->pd_len) i = P->pd_len - 1;
            Screen.PutRect (P->pd_x + 1, P->pd_y + P->pd_dy * i + 1, P->pd_x + P->pd_dx - 1, P->pd_y + P->pd_dy * i + P->pd_dy + 3, 0x0000ff);
        }
        if (P->pd_flag & Txt_AltFont) j = 2; else j = 5;
        for (i=0; i<P->pd_len; i++)
            Screen.PutTextBox (P->pd_x + j, P->pd_y + 3 + P->pd_dy * i, P->pd_x + P->pd_dx, P->pd_flag, StrTable[P->pd_base + i]);
    }
    #ifdef _DEBUG
    if (event == VE_Draw && borders) {
        Screen.Lock (Screen.back);
        for (temp = next; temp != NULL; temp = temp->next) temp->DrawBorder ();
        for (temp = link; temp != NULL; temp = temp->next) temp->DrawBorder ();
        Screen.Unlock ();
    }
    #endif
    if (domouse && P != sc_keyb) {
        MouseMove ();
        domouse = false;
    }
    if (P->mouseover) if (P->mouseover->state == VS_Visible) P->mouseover->state = VS_MouseOver;
}
